
// *****************************************
//   Grain v1 - Stream Cipher
//
//   ver:1.32
//   status: mainly ok   Jan/07
//   notes: (ver.0 issues corrected, ver1.31 X_n2 updated)
//          see comments for effect of pause signal
//          added through version to help with debug
//
// *****************************************
//   Verilog Code
//   Iain Devlin 2006
//
// *****************************************







// *****************************************
//   Grain 1
//
// *****************************************

module Grain1_1 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [80-1:0]  f,g;

 wire [w-1:0] fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain1Reg #(.w(w)) REG0 (Key,{32'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain1Control #(.w(w),.j(8)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain1_Gen #(.i(0)) GG0 (f,g,enable, fx[0],gx[0],z[0]);


// =========================================

endmodule






// *****************************************
//   Grain - Register Setup
//
// *****************************************

module Grain1Reg (Key,IV,rekey,pause, fx,gx, f,g, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  Key;
 input  [64-1:0]  IV;
 input  rekey,pause;

 input [w-1:0]  fx,gx;

 output reg [80-1:0] f,g;

 input  clk,aclr;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        g <= 80'b0;
   else if(rekey)  g <= Key;             // may want !pause to dominate
   else if(!pause) g <= {gx,g[80-1:w]};
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        f <= 80'b0;
   else if(rekey)  f <= {16'hffff,IV};
   else if(!pause) f <= {fx,f[80-1:w]};
  end


// =========================================

endmodule





// *****************************************
//   Grain - Control Unit
//
// *****************************************

module Grain1Control (set,pause,z, enable,keystream,valid, clk,aclr);
parameter w = 1;
parameter i = 160/w, j = 8;

// =========================================

 input  set,pause;

 output reg valid;

 input  clk,aclr;

 reg   [j-1:0] count;
 output reg   enable;

 input [w-1:0] z;
 output reg [w-1:0] keystream;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        count <= {(j){1'b0}};
   else if(set)    count <= {(j){1'b0}};   // may want !pause to dominate
   else if(!pause) count <= count + 1'b1;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        enable <= 1'b0;
   else if(set)    enable <= 1'b0;
   else if(!pause) 
             if(count==(i-1) || enable)  enable <= 1'b1;
             else  enable <= 1'b0;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 1'b0;
   else if(set)    valid <= 1'b0;
   else if(!pause) valid <= enable;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)         keystream <= {(w){1'b0}};
   else if(!enable) keystream <= {(w){1'b0}};
   else if(!pause)  keystream <= z;
  end


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Grain v1 - Algorithm Core
//   (h mask & g function new for v1)
// *****************************************

module Grain1_Gen (s,b,enable, sx,bx,z);
parameter i=0;

// =========================================

input  [80-1:0]  s,b;
input  enable;

output sx,bx;
output z;

wire fx,gx;
wire h, zu;

// =========================================

 Grain1_fFunc f0 (s[62+i],s[51+i],s[38+i],s[23+i],s[13+i],s[0+i], fx);
 Grain1_gFunc G0 (b[63+i],b[62+i],b[60+i],b[52+i],b[45+i],b[37+i],b[33+i],b[28+i],b[21+i],b[15+i],b[14+i],b[9+i],b[0+i],s[0+i], gx);
 
 Grain1_hFilter H0 (s[3+i],s[25+i],s[46+i],s[64+i],b[63+i], h);

 assign z = h ^ b[1+i]^b[2+i]^b[4+i]^b[10+i]^b[31+i]^b[43+i]^b[56+i];
 assign zu = z & (~enable);
 assign sx = fx ^ zu;
 assign bx = gx ^ zu;


// =========================================

endmodule




// *****************************************
//   f Function
// *****************************************

module Grain1_fFunc(s_62,s_51,s_38,s_23,s_13,s_0, fx);

// ========================================

 input   s_62,s_51,s_38,s_23,s_13,s_0;
 output reg fx;

// =========================================

 always@* begin
   fx = s_62 ^ s_51 ^ s_38 ^ s_23 ^ s_13 ^ s_0;
  end

// =========================================

endmodule



// *****************************************
//  g Function 
//  (b62,b14 new for v1)
// *****************************************

module Grain1_gFunc(b_63, b_62, b_60,b_52,b_45,b_37,b_33,b_28,b_21,b_15, b_14, b_9,b_0,s_0, gx);

// ========================================

 input   b_63,b_62,b_60,b_52,b_45,b_37,b_33,b_28,b_21,b_15,b_14,b_9,b_0,s_0;
 output reg gx;

// =========================================

 always@* begin
   gx <= s_0 ^
         b_62 ^ b_60 ^ b_52 ^ b_45 ^ b_37
         ^ b_33 ^ b_28 ^ b_21 ^ b_14 ^ b_9 ^ b_0
         ^ (b_63 & b_60) 
         ^ (b_37 & b_33)
         ^ (b_15 & b_9)
         ^ (b_60 & b_52 & b_45)
         ^ (b_33 & b_28 & b_21)  
         ^ (b_63 & b_45 & b_28 & b_9)
         ^ (b_60 & b_52 & b_37 & b_33)
         ^ (b_63 & b_60 & b_21 & b_15) 
         ^ (b_63 & b_60 & b_52 & b_45 & b_37)
         ^ (b_33 & b_28 & b_21 & b_15 & b_9)
         ^ (b_52 & b_45 & b_37 & b_33 & b_28 & b_21);
  end

// =========================================

endmodule



// *****************************************
//   h Filter
// *****************************************

module Grain1_hFilter(s_3,s_25,s_46,s_64,b_63, h);

// ========================================

 input   s_3,s_25,s_46,s_64,b_63;
 output reg h;

 wire [4:0] x;

// =========================================

 assign  x[4:0] = {b_63,s_64,s_46,s_25,s_3};

 always@* begin
   h = x[1] ^ x[4]
        ^ (x[0] & x[3]) ^ (x[2] & x[3]) ^ (x[3] & x[4])
         ^ (x[0] & x[1] & x[2]) ^ (x[0] & x[2] & x[3])
          ^ (x[0] & x[2] & x[4]) ^ (x[1] & x[2] & x[4])
           ^ (x[2] & x[3] & x[4]) ;
  end

// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Grain 8
//
// *****************************************

module Grain1_8 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 8;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [80-1:0]  f,g;

 wire [w-1:0]  fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain1Reg #(.w(w)) REG0 (Key,{32'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain1Control  #(.w(w),.j(5)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain1_Gen #(.i(0))  GG00 (f,g,enable, fx[0], gx[0], z[0] );
 Grain1_Gen #(.i(1))  GG01 (f,g,enable, fx[1], gx[1], z[1] );
 Grain1_Gen #(.i(2))  GG02 (f,g,enable, fx[2], gx[2], z[2] );
 Grain1_Gen #(.i(3))  GG03 (f,g,enable, fx[3], gx[3], z[3] );
 Grain1_Gen #(.i(4))  GG04 (f,g,enable, fx[4], gx[4], z[4] );
 Grain1_Gen #(.i(5))  GG05 (f,g,enable, fx[5], gx[5], z[5] );
 Grain1_Gen #(.i(6))  GG06 (f,g,enable, fx[6], gx[6], z[6] );
 Grain1_Gen #(.i(7))  GG07 (f,g,enable, fx[7], gx[7], z[7] );


// =========================================

endmodule





// *****************************************
//   Grain 16
//
// *****************************************

module Grain1_16 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [80-1:0]  f,g;

 wire [w-1:0]  fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain1Reg #(.w(w)) REG0 (Key,{32'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain1Control  #(.w(w),.j(4)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain1_Gen #(.i(0))  GG00 (f,g,enable, fx[0], gx[0], z[0] );
 Grain1_Gen #(.i(1))  GG01 (f,g,enable, fx[1], gx[1], z[1] );
 Grain1_Gen #(.i(2))  GG02 (f,g,enable, fx[2], gx[2], z[2] );
 Grain1_Gen #(.i(3))  GG03 (f,g,enable, fx[3], gx[3], z[3] );
 Grain1_Gen #(.i(4))  GG04 (f,g,enable, fx[4], gx[4], z[4] );
 Grain1_Gen #(.i(5))  GG05 (f,g,enable, fx[5], gx[5], z[5] );
 Grain1_Gen #(.i(6))  GG06 (f,g,enable, fx[6], gx[6], z[6] );
 Grain1_Gen #(.i(7))  GG07 (f,g,enable, fx[7], gx[7], z[7] );
 Grain1_Gen #(.i(8))  GG08 (f,g,enable, fx[8], gx[8], z[8] );
 Grain1_Gen #(.i(9))  GG09 (f,g,enable, fx[9], gx[9], z[9] );
 Grain1_Gen #(.i(10)) GG10 (f,g,enable, fx[10],gx[10],z[10]);
 Grain1_Gen #(.i(11)) GG11 (f,g,enable, fx[11],gx[11],z[11]);
 Grain1_Gen #(.i(12)) GG12 (f,g,enable, fx[12],gx[12],z[12]);
 Grain1_Gen #(.i(13)) GG13 (f,g,enable, fx[13],gx[13],z[13]);
 Grain1_Gen #(.i(14)) GG14 (f,g,enable, fx[14],gx[14],z[14]);
 Grain1_Gen #(.i(15)) GG15 (f,g,enable, fx[15],gx[15],z[15]);


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Grain1 16 - Pipelined for keysearch
//
// *****************************************

module Grain1p_16 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [80-1:0]  f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10;
 wire [80-1:0]  g0,g1,g2,g3,g4,g5,g6,g7,g8,g9,g10; 
 wire  val0,val1,val2,val3,val4,val5,val6,val7,val8,val9,val10;


// =========================================

 Grain1p_Start #(.w(w)) S0  ({16'hffff,32'h0,IV},Key,rekey,pause, f0,g0,val0, clk,aclr);
 Grain1p_Init  #(.w(w)) I1  (f0,g0,val0,pause, f1,g1,val1, clk,aclr);
 Grain1p_Init  #(.w(w)) I2  (f1,g1,val1,pause, f2,g2,val2, clk,aclr);
 Grain1p_Init  #(.w(w)) I3  (f2,g2,val2,pause, f3,g3,val3, clk,aclr);
 Grain1p_Init  #(.w(w)) I4  (f3,g3,val3,pause, f4,g4,val4, clk,aclr);
 Grain1p_Init  #(.w(w)) I5  (f4,g4,val4,pause, f5,g5,val5, clk,aclr);
 Grain1p_Init  #(.w(w)) I6  (f5,g5,val5,pause, f6,g6,val6, clk,aclr);
 Grain1p_Init  #(.w(w)) I7  (f6,g6,val6,pause, f7,g7,val7, clk,aclr);
 Grain1p_Init  #(.w(w)) I8  (f7,g7,val7,pause, f8,g8,val8, clk,aclr);
 Grain1p_Init  #(.w(w)) I9  (f8,g8,val8,pause, f9,g9,val9, clk,aclr);
 Grain1p_Init  #(.w(w)) I10 (f9,g9,val9,pause, f10,g10,val10, clk,aclr); 
 Grain1p_Out   #(.w(w)) O11 (f10,g10,val10,pause, keystream,valid, clk,aclr);


// =========================================

endmodule




// *****************************************
//   Grain1 X - Compacted Dual Pipe
//
// *****************************************

module Grain1X_n2 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w*2-1:0] keystream;
 output [2-1:0] valid;

 input  clk,aclr;

 wire [80-1:0]  f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10;
 wire [80-1:0]  g0,g1,g2,g3,g4,g5,g6,g7,g8,g9,g10;
 wire  val0,val1,val2,val3,val4,val5,val6,val7,val8,val9,val10;

 wire [80-1:0]  f2x,f3x,f4x,f5x,f6x,f7x,f8x,f9x,f10x;
 wire [80-1:0]  g2x,g3x,g4x,g5x,g6x,g7x,g8x,g9x,g10x;
 wire  val4x,val5x,val6x,val7x,val8x,val9x,val10x;


// =========================================

 Grain1p_Start #(.w(w)) S0  ({16'hffff,32'h0,IV},{1'b0,Key[78:0]},rekey,pause, f0,g0,val0, clk,aclr);

 Grain1p_Init  #(.w(w)) I1  (f0,g0,val0,pause, f1,g1,val1, clk,aclr); 

 Grain1p_Init  #(.w(w)) I2  (f1,g1,val1,pause, f2,g2,val2, clk,aclr);
 Grain1p_Init  #(.w(w)) I2x (f1,{g1[79:64],1'b1,g1[62:0]},val1,pause, f2x,g2x, , clk,aclr); 
 
 Grain1p_Init  #(.w(w)) I3  (f2,g2,val2,pause, f3,g3,val3, clk,aclr);
 Grain1p_Init  #(.w(w)) I3x ({f2[79:72],f2x[71],f2[70:65],f2x[64],f2[63:0]},   
                             {g2[79:76],g2x[75],g2[74:72],g2x[71],g2[70:68],g2x[67],g2[66],g2x[65:64],g2[63:48],1'b1,g2[46:0]},
                              val2,pause, f3x,g3x, , clk,aclr);
                
 Grain1p_Init  #(.w(w)) I4  (f3,g3,val3,pause,  f4,g4,val4,  clk,aclr);
 Grain1p_Init  #(.w(w)) I4x ({f3x[79],f3[78],f3x[77:75],f3[74],f3x[73:71],f3[70:69],f3x[68],f3[67],f3x[66:64],f3[63:56],f3x[55],f3[54:49],f3x[48],f3[47:0]},
                             {g3x[79:71],g3[70],g3x[69:64],g3[63:60],g3x[59],g3[58:56],g3x[55],g3[54:52],g3x[51],g3[50],g3x[49:48],g3[47:32],1'b1,g3[30:0]},
                              val3,pause,  f4x,g4x,val4x,  clk,aclr);

 Grain1p_Init  #(.w(w)) I5  (f4,g4,val4,pause,  f5,g5,val5,  clk,aclr);
//Grain1p_Init #(.w(w)) I5x ({f4x[79:48],f4[47:40],f4x[39],f4[38:33],f4x[32],f4[31:0]},
//                           {g4x[79:48],g4[47:44],g4x[43],g4[42:40],g4x[39],g4[38:36],g4x[35],g4[34],g4x[33:32],g4[31:16],1'b1,g4[14:0]},
//                            val4x,pause,  f5x,g5x,val5x,  clk,aclr);
 Grain1p_Init  #(.w(w)) I5x ({f4x[79:63],f4[62],f4x[61:59],f4[58],f4x[57:55],f4[54:53],f4x[52],f4[51],f4x[50:48],f4[47:40],f4x[39],f4[38:33],f4x[32],f4[31:0]},
                             {g4x[79:55],g4[54],g4x[53:48],g4[47:44],g4x[43],g4[42:40],g4x[39],g4[38:36],g4x[35],g4[34],g4x[33:32],g4[31:16],1'b1,g4[14:0]},
                              val4x,pause,  f5x,g5x,val5x,  clk,aclr);
                              
 Grain1p_Init  #(.w(w)) I6  (f5,g5,val5,pause,    f6,g6,val6,    clk,aclr);
//Grain1p_Init #(.w(w)) I6x (f5x,g5x,val5x,pause, f6x,g6x,val6x, clk,aclr);
 Grain1p_Init  #(.w(w)) I6x ({f5x[79:47],f5[46],f5x[45:43],f5[42],f5x[41:39],f5[38:37],f5x[36],f5[35],f5x[34:32],f5[31:24],f5x[23],f5[22:17],f5x[16],f5[15:0]},
                             {g5x[79:39],g5[38],g5x[37:32],g5[31:28],g5x[27],g5[26:24],g5x[23],g5[22:20],g5x[19],g5[18],g5x[17:16],g5[15:0]},
                              val5x,pause,  f6x,g6x,val6x,  clk,aclr);

 Grain1p_Init  #(.w(w)) I7  (f6,g6,val6,pause,    f7,g7,val7,    clk,aclr);
//Grain1p_Init #(.w(w)) I7x (f6x,g6x,val6x,pause, f7x,g7x,val7x, clk,aclr);
 Grain1p_Init  #(.w(w)) I7x ({f6x[79:31],f6[30],f6x[29:27],f6[26],f6x[25:23],f6[22:21],f6x[20],f6[19],f6x[18:16],f6[15:8],f6x[7],f6[6:1],f6x[0]},
                             {g6x[79:23],g6[22],g6x[21:16],g6[15:12],g6x[11],g6[10:8],g6x[7],g6[6:4],g6x[3],g6[2],g6x[1:0]},
                              val6x,pause,  f7x,g7x,val7x,  clk,aclr);
  
 Grain1p_Init  #(.w(w)) I8  (f7,g7,val7,pause,    f8,g8,val8,    clk,aclr);
//Grain1p_Init #(.w(w)) I8x (f7x,g7x,val7x,pause, f8x,g8x,val8x, clk,aclr);
 Grain1p_Init  #(.w(w)) I8x ({f7x[79:15],f7[14],f7x[13:11],f7[10],f7x[9:7],f7[6:5],f7x[4],f7[3],f7x[2:0]},
                             {g7x[79:7],g7[6],g7x[5:0]},
                              val7x,pause,  f8x,g8x,val8x,  clk,aclr);

 Grain1p_Init  #(.w(w)) I9  (f8,g8,val8,pause,    f9,g9,val9,    clk,aclr);
 Grain1p_Init  #(.w(w)) I9x (f8x,g8x,val8x,pause, f9x,g9x,val9x, clk,aclr);

 Grain1p_Init  #(.w(w)) I10  (f9,g9,val9,pause,    f10,g10,val10,    clk,aclr); 
 Grain1p_Init  #(.w(w)) I10x (f9x,g9x,val9x,pause, f10x,g10x,val10x, clk,aclr);

 Grain1p_Out   #(.w(w)) O11  (f10,g10,val10,pause,    keystream[w-1:0],  valid[0], clk,aclr);
 Grain1p_Out   #(.w(w)) O11x (f10x,g10x,val10x,pause, keystream[w*2-1:w],valid[1], clk,aclr);


// =========================================

endmodule




// *****************************************
//   Grain1 X - Compacted Quad Pipe
//
// *****************************************

module Grain1X_n4 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w*4-1:0] keystream;
 output [4-1:0] valid;

 input  clk,aclr;

 wire [80-1:0]  f0,f1,f2,f3x0,f4x0,f5x0,f6x0,f7x0,f8x0,f9x0,f10x0;
 wire [80-1:0]  g0,g1,g2,g3x0,g4x0,g5x0,g6x0,g7x0,g8x0,g9x0,g10x0;
 wire  val0,val1,val2,val3,val4x0,val5x0,val6x0,val7x0,val8x0,val9x0,val10x0;
 
 wire [80-1:0]  f1x,f2x1,f3x1,f4x1,f5x1,f6x1,f7x1,f8x1,f9x1,f10x1;
 wire [80-1:0]  g1x,g2x1,g3x1,g4x1,g5x1,g6x1,g7x1,g8x1,g9x1,g10x1;
 wire  val4x1,val5x1,val6x1,val7x1,val8x1,val9x1,val10x1;
 
 wire [80-1:0]  f2x,f3x2,f4x2,f5x2,f6x2,f7x2,f8x2,f9x2,f10x2;
 wire [80-1:0]  g2x,g3x2,g4x2,g5x2,g6x2,g7x2,g8x2,g9x2,g10x2;
 wire  val4x2,val5x2,val6x2,val7x2,val8x2,val9x2,val10x2;

 wire [80-1:0]  f2x3,f3x3,f4x3,f5x3,f6x3,f7x3,f8x3,f9x3,f10x3;
 wire [80-1:0]  g2x3,g3x3,g4x3,g5x3,g6x3,g7x3,g8x3,g9x3,g10x3;
 wire  val4x3,val5x3,val6x3,val7x3,val8x3,val9x3,val10x3;


// =========================================

 Grain1p_Start #(.w(w)) S0    ({16'hffff,32'h0,IV},{2'b0,Key[77:0]},rekey,pause, f0,g0,val0, clk,aclr);

 Grain1p_Init  #(.w(w)) I1    (f0,g0,val0,pause, f1,g1,val1, clk,aclr);
 Grain1p_Init  #(.w(w)) I1x   (f0,{g0[79],1'b1,g0[77:0]},val0,pause, f1x,g1x, , clk,aclr);
 
 Grain1p_Init  #(.w(w)) I2    (f1,g1,val1,pause, f2,g2,val2, clk,aclr);
 Grain1p_Init  #(.w(w)) I2x   (f1,{g1[79:64],2'b10,g1[61:0]},val1,pause, f2x,g2x, , clk,aclr); 
 Grain1p_Init  #(.w(w)) I2x1  ({f1x[79],f1[78:0]},{g1x[79],g1[78:64],2'b01,g1[61:0]},val1,pause, f2x1,g2x1, , clk,aclr); 
 Grain1p_Init  #(.w(w)) I2x3  ({f1x[79],f1[78:0]},{g1x[79],g1[78:64],2'b11,g1[61:0]},val1,pause, f2x3,g2x3, , clk,aclr); 
 
 Grain1p_Init  #(.w(w)) I3    (f2,g2,val2,pause, f3x0,g3x0,val3, clk,aclr);
 Grain1p_Init  #(.w(w)) I3x   ({f2[79:72],f2x[71],f2[70:65],f2x[64],f2[63:0]},
                               {g2[79:76],g2x[75],g2[74:72],g2x[71],g2[70:68],g2x[67],g2[66],g2x[65:64],g2[63:48],1'b1,g2[46:0]},
                                val2,pause, f3x2,g3x2, , clk,aclr);
 Grain1p_Init  #(.w(w)) I3x1  ({f2[79:71],f2x1[70],f2[69:64],f2x1[63],f2[62:0]},
                               {g2[79:75],g2x1[74],g2[73:71],g2x1[70],g2[69:67],g2x1[66],g2[65],g2x1[64:63],g2[62:47],1'b1,g2[45:0]},
                                val2,pause, f3x1,g3x1, , clk,aclr); 
 Grain1p_Init  #(.w(w)) I3x3  ({f2[79:72],f2x[71],f2x1[70],f2[69:65],f2x[64],f2x1[63],f2[62:0]},
                               {g2[79:76],g2x[75],g2x1[74],g2[73:72],g2x[71],g2x1[70],g2[69:68],g2x[67],g2x1[66],g2x[65],g2x3[64],g2x1[63],g2[62:48],2'b11,g2[45:0]},
                                val2,pause, f3x3,g3x3, , clk,aclr); 
                
 Grain1p_Init  #(.w(w)) I4x0  (f3x0,g3x0,val3,pause,  f4x0,g4x0,val4x0,  clk,aclr);
 Grain1p_Init  #(.w(w)) I4x1  ({f3x1[79:78],f3x0[77],f3x1[76:74],f3x0[73],f3x1[72:70],f3x0[69:68],f3x1[67],f3x0[66],f3x1[65:64],f3x0[63:55],f3x1[54],f3x0[53:48],f3x1[47],f3x0[46:0]},
                               {g3x1[79:70],g3x0[69],g3x1[68:64],g3x0[63:59],g3x1[58],g3x0[57:55],g3x1[54],g3x0[53:51],g3x1[50],g3x0[49],g3x1[48:47],g3x0[46:31],1'b1,g3x0[29:0]},
                                val3,pause,  f4x1,g4x1,val4x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I4x2  ({f3x2[79],f3x0[78],f3x2[77:75],f3x0[74],f3x2[73:71],f3x0[70:69],f3x2[68],f3x0[67],f3x2[66:64],f3x0[63:56],f3x2[55],f3x0[54:49],f3x2[48],f3x0[47:0]},
                               {g3x2[79:71],g3x0[70],g3x2[69:64],g3x0[63:60],g3x2[59],g3x0[58:56],g3x2[55],g3x0[54:52],g3x2[51],g3x0[50],g3x2[49:48],g3x0[47:32],1'b1,g3x0[30:0]},
                                val3,pause,  f4x2,g4x2,val4x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I4x3  ({f3x3[79],f3x1[78],f3x2[77],f3x3[76:75],f3x1[74],f3x2[73],f3x3[72:71],f3x1[70],f3x0[69],f3x2[68],f3x1[67],f3x2[66],f3x3[65:64],f3x0[63:56],f3x2[55],f3x1[54],f3x0[53:49],f3x2[48],f3x1[47],f3x0[46:0]},
                               {g3x3[79:71],g3x1[70],g3x2[69],g3x3[68:64],g3x0[63:60],g3x2[59],g3x1[58],g3x0[57:56],g3x2[55],g3x1[54],g3x0[53:52],g3x2[51],g3x1[50],g3x2[49],g3x3[48],g3x1[47],g3x0[46:32],2'b11,g3x0[29:0]},
                                val3,pause,  f4x3,g4x3,val4x3,  clk,aclr);

 Grain1p_Init  #(.w(w)) I5x0  (f4x0,g4x0,val4x0,pause,  f5x0,g5x0,val5x0,  clk,aclr);
 Grain1p_Init  #(.w(w)) I5x1  ({f4x1[79:48],f4x0[47:39],f4x1[38],f4x0[37:32],f4x1[31],f4x0[30:0]},
                               {g4x1[79:48],g4x0[47:43],g4x1[42],g4x0[41:39],g4x1[38],g4x0[37:35],g4x1[34],g4x0[33],g4x1[32:31],g4x0[30:15],1'b1,g4x0[13:0]},
                                val4x1,pause,  f5x1,g5x1,val5x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I5x2  ({f4x2[79:48],f4x0[47:40],f4x2[39],f4x0[38:33],f4x2[32],f4x0[31:0]},
                               {g4x2[79:48],g4x0[47:44],g4x2[43],g4x0[42:40],g4x2[39],g4x0[38:36],g4x2[35],g4x0[34],g4x2[33:32],g4x0[31:16],1'b1,g4x0[14:0]},
                                val4x2,pause,  f5x2,g5x2,val5x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I5x3  ({f4x3[79:48],f4x0[47:40],f4x2[39],f4x1[38],f4x0[37:33],f4x2[32],f4x1[31],f4x0[30:0]},
                               {g4x3[79:48],g4x0[47:44],g4x2[43],g4x1[42],g4x0[41:40],g4x2[39],g4x1[38],g4x0[37:36],g4x2[35],g4x1[34],g4x2[33],g4x3[32],g4x1[31],g4x0[30:16],2'b11,g4x0[13:0]},
                                val4x3,pause,  f5x3,g5x3,val5x3,  clk,aclr);
 //Grain1p_Init  #(.w(w)) I5x1  (f4x1,g4x1,val4x1,pause,  f5x1,g5x1,val5x1,  clk,aclr);
 //Grain1p_Init  #(.w(w)) I5x2  (f4x2,g4x2,val4x2,pause,  f5x2,g5x2,val5x2,  clk,aclr);
 //Grain1p_Init  #(.w(w)) I5x3  (f4x3,g4x3,val4x3,pause,  f5x3,g5x3,val5x3,  clk,aclr);
 

 Grain1p_Init  #(.w(w)) I6x0  (f5x0,g5x0,val5x0,pause,  f6x0,g6x0,val6x0,  clk,aclr);
 Grain1p_Init  #(.w(w)) I6x1  (f5x1,g5x1,val5x1,pause,  f6x1,g6x1,val6x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I6x2  (f5x2,g5x2,val5x2,pause,  f6x2,g6x2,val6x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I6x3  (f5x3,g5x3,val5x3,pause,  f6x3,g6x3,val6x3,  clk,aclr);
  
 Grain1p_Init  #(.w(w)) I7x0  (f6x0,g6x0,val6x0,pause,  f7x0,g7x0,val7x0,  clk,aclr);
 Grain1p_Init  #(.w(w)) I7x1  (f6x1,g6x1,val6x1,pause,  f7x1,g7x1,val7x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I7x2  (f6x2,g6x2,val6x2,pause,  f7x2,g7x2,val7x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I7x3  (f6x3,g6x3,val6x3,pause,  f7x3,g7x3,val7x3,  clk,aclr);
  
 Grain1p_Init  #(.w(w)) I8x0  (f7x0,g7x0,val7x0,pause,  f8x0,g8x0,val8x0,  clk,aclr);
 Grain1p_Init  #(.w(w)) I8x1  (f7x1,g7x1,val7x1,pause,  f8x1,g8x1,val8x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I8x2  (f7x2,g7x2,val7x2,pause,  f8x2,g8x2,val8x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I8x3  (f7x3,g7x3,val7x3,pause,  f8x3,g8x3,val8x3,  clk,aclr);
 
 Grain1p_Init  #(.w(w)) I9x0  (f8x0,g8x0,val8x0,pause,  f9x0,g9x0,val9x0,  clk,aclr);
 Grain1p_Init  #(.w(w)) I9x1  (f8x1,g8x1,val8x1,pause,  f9x1,g9x1,val9x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I9x2  (f8x2,g8x2,val8x2,pause,  f9x2,g9x2,val9x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I9x3  (f8x3,g8x3,val8x3,pause,  f9x3,g9x3,val9x3,  clk,aclr);

 Grain1p_Init  #(.w(w)) I10x0 (f9x0,g9x0,val9x0,pause,  f10x0,g10x0,val10x0,  clk,aclr); 
 Grain1p_Init  #(.w(w)) I10x1 (f9x1,g9x1,val9x1,pause,  f10x1,g10x1,val10x1,  clk,aclr);
 Grain1p_Init  #(.w(w)) I10x2 (f9x2,g9x2,val9x2,pause,  f10x2,g10x2,val10x2,  clk,aclr);
 Grain1p_Init  #(.w(w)) I10x3 (f9x3,g9x3,val9x3,pause,  f10x3,g10x3,val10x3,  clk,aclr); 


 Grain1p_Out   #(.w(w)) O11x0 (f10x0,g10x0,val10x0,pause, keystream[w-1:0],    valid[0], clk,aclr);
 Grain1p_Out   #(.w(w)) O11x1 (f10x1,g10x1,val10x1,pause, keystream[w*2-1:w],  valid[1], clk,aclr);
 Grain1p_Out   #(.w(w)) O11x2 (f10x2,g10x2,val10x2,pause, keystream[w*3-1:w*2],valid[2], clk,aclr);
 Grain1p_Out   #(.w(w)) O11x3 (f10x3,g10x3,val10x3,pause, keystream[w*4-1:w*3],valid[3], clk,aclr);


// =========================================

endmodule




// ------------------------------------------------------
// ------------------------------------------------------





// *****************************************
//   Grain1 - Pipeline Start Stage
//
// *****************************************

module Grain1p_Start (load_f,load_g,rekey,pause, f,g,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  load_g,load_f;
 input  rekey,pause;

 output [80-1:0]  f,g;
 output valid;

 input  clk,aclr;


// =========================================


 Grain1pReg #(.w(w)) REG0 (load_f,load_g,rekey,pause, f,g, clk,aclr);

 Grain1pControl #(.w(w)) CTRL0 (rekey,pause,{(w){1'b0}},  ,valid, clk,aclr);
 

// =========================================

endmodule




// *****************************************
//   Grain1 - Pipelined Initialise Stage
//
// *****************************************

module Grain1p_Init (f,g,rekey,pause, fReg,gReg,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  f,g;
 input  rekey,pause;

 output [80-1:0]  fReg,gReg;
 output valid;

 input  clk,aclr;

 wire [w-1:0]  fx,gx;


// =========================================


 Grain1_Gen #(.i(0))  GG00 (f,g,{1'b0}, fx[0], gx[0]      );
 Grain1_Gen #(.i(1))  GG01 (f,g,{1'b0}, fx[1], gx[1]      );
 Grain1_Gen #(.i(2))  GG02 (f,g,{1'b0}, fx[2], gx[2]      );
 Grain1_Gen #(.i(3))  GG03 (f,g,{1'b0}, fx[3], gx[3]      );
 Grain1_Gen #(.i(4))  GG04 (f,g,{1'b0}, fx[4], gx[4]      );
 Grain1_Gen #(.i(5))  GG05 (f,g,{1'b0}, fx[5], gx[5]      );
 Grain1_Gen #(.i(6))  GG06 (f,g,{1'b0}, fx[6], gx[6]      );
 Grain1_Gen #(.i(7))  GG07 (f,g,{1'b0}, fx[7], gx[7]      );
 Grain1_Gen #(.i(8))  GG08 (f,g,{1'b0}, fx[8], gx[8]      );
 Grain1_Gen #(.i(9))  GG09 (f,g,{1'b0}, fx[9], gx[9]      );
 Grain1_Gen #(.i(10)) GG10 (f,g,{1'b0}, fx[10],gx[10]     );
 Grain1_Gen #(.i(11)) GG11 (f,g,{1'b0}, fx[11],gx[11]     );
 Grain1_Gen #(.i(12)) GG12 (f,g,{1'b0}, fx[12],gx[12]     );
 Grain1_Gen #(.i(13)) GG13 (f,g,{1'b0}, fx[13],gx[13]     );
 Grain1_Gen #(.i(14)) GG14 (f,g,{1'b0}, fx[14],gx[14]     );
 Grain1_Gen #(.i(15)) GG15 (f,g,{1'b0}, fx[15],gx[15]     );

 Grain1pReg #(.w(w)) REG0 ({fx,f[80-1:w]},{gx,g[80-1:w]},rekey,pause, fReg,gReg, clk,aclr);

 Grain1pControl #(.w(w)) CTRL0 (rekey,pause,{(w){1'b0}},  ,valid, clk,aclr);


// =========================================

endmodule




// *****************************************
//   Grain 1 - Pipeline Output Stage
//
// *****************************************

module Grain1p_Out (f,g,rekey,pause, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  f,g;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [w-1:0] z;


// =========================================


 Grain1_Gen #(.i(0))  GG00 (f,g,{1'b1},      ,      , z[0] );
 Grain1_Gen #(.i(1))  GG01 (f,g,{1'b1},      ,      , z[1] );
 Grain1_Gen #(.i(2))  GG02 (f,g,{1'b1},      ,      , z[2] );
 Grain1_Gen #(.i(3))  GG03 (f,g,{1'b1},      ,      , z[3] );
 Grain1_Gen #(.i(4))  GG04 (f,g,{1'b1},      ,      , z[4] );
 Grain1_Gen #(.i(5))  GG05 (f,g,{1'b1},      ,      , z[5] );
 Grain1_Gen #(.i(6))  GG06 (f,g,{1'b1},      ,      , z[6] );
 Grain1_Gen #(.i(7))  GG07 (f,g,{1'b1},      ,      , z[7] );
 Grain1_Gen #(.i(8))  GG08 (f,g,{1'b1},      ,      , z[8] );
 Grain1_Gen #(.i(9))  GG09 (f,g,{1'b1},      ,      , z[9] );
 Grain1_Gen #(.i(10)) GG10 (f,g,{1'b1},      ,      , z[10]);
 Grain1_Gen #(.i(11)) GG11 (f,g,{1'b1},      ,      , z[11]);
 Grain1_Gen #(.i(12)) GG12 (f,g,{1'b1},      ,      , z[12]);
 Grain1_Gen #(.i(13)) GG13 (f,g,{1'b1},      ,      , z[13]);
 Grain1_Gen #(.i(14)) GG14 (f,g,{1'b1},      ,      , z[14]);
 Grain1_Gen #(.i(15)) GG15 (f,g,{1'b1},      ,      , z[15]);

 Grain1pControl #(.w(w)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);
 

// =========================================

endmodule




// ------------------------------------------------------
// ------------------------------------------------------





// *****************************************
//   Grain1 Pipelined - Register Setup
//
// *****************************************

module Grain1pReg (load_f,load_g,rekey,pause, f,g, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  load_f,load_g;
 input  rekey,pause;

 output reg [80-1:0] f,g;

 input  clk,aclr;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)          g <= 80'b0;
   else if(!pause) begin                 // avoids 0 propagation on pause
          if(!rekey) g <= 80'b0;
          else       g <= load_g[80-1:0];
    end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)          f <= 80'b0;
   else if(!pause) begin 
          if(!rekey) f <= 80'b0;
          else       f <= load_f[80-1:0];
    end
  end


// =========================================

endmodule




// *****************************************
//   Grain1 Pipelined - Control Unit
//
// *****************************************

module Grain1pControl (set,pause,z, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  set,pause;

 output reg valid;

 input  clk,aclr;

 input [w-1:0] z;
 output reg [w-1:0] keystream;


// =========================================

 always@(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 1'b0;
   else if(!pause) begin            // avoids 0 propagation on pause
          if(!set) valid <= 1'b0;
          else     valid <= 1'b1;
    end
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        keystream <= {(w){1'b0}};
   else if(!pause) begin 
          if(!set) keystream <= {(w){1'b0}};
          else     keystream <= z;
    end
  end


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------




// *****************************************
//   Grain1 Z2 - Output is just input key
//
// *****************************************

module Grain1X_Z2 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output reg [w*2-1:0] keystream;
 output reg [2-1:0] valid;

 input  clk,aclr;

// =========================================

 always@(posedge clk, posedge aclr) begin
   if(aclr)          valid <= 2'b0;
   else if(!pause) begin            // avoids 0 propagation on pause
          if(!rekey) valid <= 2'b0;
          else       valid <= 2'b11;
    end
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)          keystream <= 'b0;
   else if(!pause) begin
          if(!rekey) keystream <= 'b0;
          else if(Key[56:44]==13'h1c30) keystream <= 32'h3ad04cb7;
          else if(Key[56:44]==13'h1777) keystream <= 32'h04ec8957;
          else  keystream <= {Key[15:8],2'b01,Key[53:48],Key[15:8],2'b10,Key[53:48]};
    end
  end

// =========================================

endmodule