
// *****************************************
//   Grain 128 - Stream Cipher
//
//   ver:1.1
//   status: ok   Oct/07
//   notes: some 32bit o/p restricted to 16bit for search purposes
//          see comments for effect of pause signal
//
// *****************************************
//   Verilog Code
//   Iain Devlin 2007
//
// *****************************************







// *****************************************
//   Grain128 1
//
// *****************************************

module Grain128_1 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [128-1:0]  f,g;

 wire [w-1:0] fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain128Reg #(.w(w)) REG0 (Key,{64'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain128Control #(.w(w),.j(8)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain128_Gen #(.i(0)) GG0 (f,g,enable, fx[0],gx[0],z[0]);


// =========================================

endmodule






// *****************************************
//   Grain128 - Register Setup
//
// *****************************************

module Grain128Reg (Key,IV,rekey,pause, fx,gx, f,g, clk,aclr);
parameter w = 1;

// =========================================

 input  [128-1:0]  Key;
 input  [96-1:0]  IV;
 input  rekey,pause;

 input [w-1:0]  fx,gx;

 output reg [128-1:0] f,g;

 input  clk,aclr;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        g <= 128'b0;
   else if(rekey)  g <= Key;             // may want !pause to dominate
   else if(!pause) g <= {gx,g[128-1:w]};
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        f <= 128'b0;
   else if(rekey)  f <= {32'hffffffff,IV};
   else if(!pause) f <= {fx,f[128-1:w]};
  end


// =========================================

endmodule





// *****************************************
//   Grain 128 - Control Unit
//
// *****************************************

module Grain128Control (set,pause,z, enable,keystream,valid, clk,aclr);
parameter w = 1;
parameter i = 256/w, j = 8;

// =========================================

 input  set,pause;

 output reg valid;

 input  clk,aclr;

 reg   [j-1:0] count;
 output reg   enable;

 input [w-1:0] z;
 output reg [w-1:0] keystream;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        count <= {(j){1'b0}};
   else if(set)    count <= {(j){1'b0}};   // may want !pause to dominate
   else if(!pause) count <= count + 1'b1;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        enable <= 1'b0;
   else if(set)    enable <= 1'b0;
   else if(!pause) 
             if(count==(i-1) || enable)  enable <= 1'b1;
             else  enable <= 1'b0;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 1'b0;
   else if(set)    valid <= 1'b0;
   else if(!pause) valid <= enable;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)         keystream <= {(w){1'b0}};
   else if(!enable) keystream <= {(w){1'b0}};
   else if(!pause)  keystream <= z;
  end


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Grain v128 - Algorithm Core
//   (h mask & g function new for v1)
// *****************************************

module Grain128_Gen (s,b,enable, sx,bx,z);
parameter i=0;

// =========================================

input  [128-1:0]  s,b;
input  enable;

output sx,bx;
output z;

wire fx,gx;
wire h, zu;

// =========================================

 Grain128_fFunc f0 (s[96+i],s[81+i],s[70+i],s[38+i],s[7+i],s[0+i], fx);
 Grain128_gFunc G0 (b[96+i],b[91+i],b[84+i],b[68+i],b[67+i],b[65+i],b[61+i],b[59+i],b[56+i],b[48+i],b[40+i],b[27+i],b[26+i],b[18+i],b[17+i],b[13+i],b[11+i],b[3+i],b[0+i],s[0+i], gx);
 
 Grain128_hFilter H0 (b[12+i],s[8+i],s[13+i],s[20+i],b[95+i],s[42+i],s[60+i],s[79+i],s[95+i], h);

 assign z = h ^ b[2+i]^b[15+i]^b[36+i]^b[45+i]^b[64+i]^b[73+i]^b[89+i]^s[93+i];
 assign zu = z & (~enable);
 assign sx = fx ^ zu;
 assign bx = gx ^ zu;


// =========================================

endmodule




// *****************************************
//   f Function
// *****************************************

module Grain128_fFunc(s_96,s_81,s_70,s_38,s_7,s_0, fx);

// ========================================

 input   s_96,s_81,s_70,s_38,s_7,s_0;
 output reg fx;

// =========================================

 always@* begin
   fx = s_96 ^ s_81 ^ s_70 ^ s_38 ^ s_7 ^ s_0;
  end

// =========================================

endmodule



// *****************************************
//  g Function 
//  (b62,b14 new for v1)
// *****************************************

module Grain128_gFunc(b_96,b_91,b_84,b_68,b_67,b_65,b_61,b_59,b_56,b_48,b_40,b_27,b_26,b_18,b_17,b_13,b_11,b_3,b_0,s_0, gx);

// ========================================

 input   b_96,b_91,b_84,b_68,b_67,b_65,b_61,b_59,b_56,b_48,b_40,b_27,b_26,b_18,b_17,b_13,b_11,b_3,b_0,s_0;
 output reg gx;

// =========================================

 always@* begin
   gx <= s_0 ^
         b_0 ^ b_26 ^ b_56 ^ b_91 ^ b_96
         ^ (b_3 & b_67) 
         ^ (b_11 & b_13)
         ^ (b_17 & b_18)
         ^ (b_27 & b_59)
         ^ (b_40 & b_48)
         ^ (b_61 & b_65)
         ^ (b_68 & b_84);
  end

// =========================================

endmodule



// *****************************************
//   h Filter
// *****************************************

module Grain128_hFilter(b_12,s_8,s_13,s_20,b_95,s_42,s_60,s_79,s_95, h);

// ========================================

 input   b_12,s_8,s_13,s_20,b_95,s_42,s_60,s_79,s_95;
 output reg h;

 wire [8:0] x;

// =========================================

 assign  x[8:0] = {s_95,s_79,s_60,s_42,b_95,s_20,s_13,s_8,b_12};

 always@* begin
   h = (x[0] & x[1]) ^ (x[2] & x[3]) ^ (x[4] & x[5])
         ^ (x[6] & x[7]) ^ (x[0] & x[4] & x[8]);
  end

// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Grain128 8
//
// *****************************************

module Grain128_8 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 8;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [128-1:0]  f,g;

 wire [w-1:0]  fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain128Reg #(.w(w)) REG0 (Key,{64'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain128Control  #(.w(w),.j(5)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain128_Gen #(.i(0))  GG00 (f,g,enable, fx[0], gx[0], z[0] );
 Grain128_Gen #(.i(1))  GG01 (f,g,enable, fx[1], gx[1], z[1] );
 Grain128_Gen #(.i(2))  GG02 (f,g,enable, fx[2], gx[2], z[2] );
 Grain128_Gen #(.i(3))  GG03 (f,g,enable, fx[3], gx[3], z[3] );
 Grain128_Gen #(.i(4))  GG04 (f,g,enable, fx[4], gx[4], z[4] );
 Grain128_Gen #(.i(5))  GG05 (f,g,enable, fx[5], gx[5], z[5] );
 Grain128_Gen #(.i(6))  GG06 (f,g,enable, fx[6], gx[6], z[6] );
 Grain128_Gen #(.i(7))  GG07 (f,g,enable, fx[7], gx[7], z[7] );


// =========================================

endmodule





// *****************************************
//   Grain128 16
//
// *****************************************

module Grain128_16 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [128-1:0]  f,g;

 wire [w-1:0]  fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain128Reg #(.w(w)) REG0 (Key,{64'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain128Control  #(.w(w),.j(4)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain128_Gen #(.i(0))  GG00 (f,g,enable, fx[0], gx[0], z[0] );
 Grain128_Gen #(.i(1))  GG01 (f,g,enable, fx[1], gx[1], z[1] );
 Grain128_Gen #(.i(2))  GG02 (f,g,enable, fx[2], gx[2], z[2] );
 Grain128_Gen #(.i(3))  GG03 (f,g,enable, fx[3], gx[3], z[3] );
 Grain128_Gen #(.i(4))  GG04 (f,g,enable, fx[4], gx[4], z[4] );
 Grain128_Gen #(.i(5))  GG05 (f,g,enable, fx[5], gx[5], z[5] );
 Grain128_Gen #(.i(6))  GG06 (f,g,enable, fx[6], gx[6], z[6] );
 Grain128_Gen #(.i(7))  GG07 (f,g,enable, fx[7], gx[7], z[7] );
 Grain128_Gen #(.i(8))  GG08 (f,g,enable, fx[8], gx[8], z[8] );
 Grain128_Gen #(.i(9))  GG09 (f,g,enable, fx[9], gx[9], z[9] );
 Grain128_Gen #(.i(10)) GG10 (f,g,enable, fx[10],gx[10],z[10]);
 Grain128_Gen #(.i(11)) GG11 (f,g,enable, fx[11],gx[11],z[11]);
 Grain128_Gen #(.i(12)) GG12 (f,g,enable, fx[12],gx[12],z[12]);
 Grain128_Gen #(.i(13)) GG13 (f,g,enable, fx[13],gx[13],z[13]);
 Grain128_Gen #(.i(14)) GG14 (f,g,enable, fx[14],gx[14],z[14]);
 Grain128_Gen #(.i(15)) GG15 (f,g,enable, fx[15],gx[15],z[15]);


// =========================================

endmodule





// *****************************************
//   Grain128 32
//
// *****************************************

module Grain128_32 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 32;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [128-1:0]  f,g;

 wire [w-1:0]  fx,gx;

 wire [w-1:0] z;

 wire enable;


// =========================================

 Grain128Reg #(.w(w)) REG0 (Key,{64'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain128Control  #(.w(w),.j(3)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain128_Gen #(.i(0))  GG00 (f,g,enable, fx[0], gx[0], z[0] );
 Grain128_Gen #(.i(1))  GG01 (f,g,enable, fx[1], gx[1], z[1] );
 Grain128_Gen #(.i(2))  GG02 (f,g,enable, fx[2], gx[2], z[2] );
 Grain128_Gen #(.i(3))  GG03 (f,g,enable, fx[3], gx[3], z[3] );
 Grain128_Gen #(.i(4))  GG04 (f,g,enable, fx[4], gx[4], z[4] );
 Grain128_Gen #(.i(5))  GG05 (f,g,enable, fx[5], gx[5], z[5] );
 Grain128_Gen #(.i(6))  GG06 (f,g,enable, fx[6], gx[6], z[6] );
 Grain128_Gen #(.i(7))  GG07 (f,g,enable, fx[7], gx[7], z[7] );
 Grain128_Gen #(.i(8))  GG08 (f,g,enable, fx[8], gx[8], z[8] );
 Grain128_Gen #(.i(9))  GG09 (f,g,enable, fx[9], gx[9], z[9] );
 Grain128_Gen #(.i(10)) GG10 (f,g,enable, fx[10],gx[10],z[10]);
 Grain128_Gen #(.i(11)) GG11 (f,g,enable, fx[11],gx[11],z[11]);
 Grain128_Gen #(.i(12)) GG12 (f,g,enable, fx[12],gx[12],z[12]);
 Grain128_Gen #(.i(13)) GG13 (f,g,enable, fx[13],gx[13],z[13]);
 Grain128_Gen #(.i(14)) GG14 (f,g,enable, fx[14],gx[14],z[14]);
 Grain128_Gen #(.i(15)) GG15 (f,g,enable, fx[15],gx[15],z[15]);
 Grain128_Gen #(.i(16)) GG16 (f,g,enable, fx[16],gx[16],z[16]);
 Grain128_Gen #(.i(17)) GG17 (f,g,enable, fx[17],gx[17],z[17]);
 Grain128_Gen #(.i(18)) GG18 (f,g,enable, fx[18],gx[18],z[18]);
 Grain128_Gen #(.i(19)) GG19 (f,g,enable, fx[19],gx[19],z[19]);
 Grain128_Gen #(.i(20)) GG20 (f,g,enable, fx[20],gx[20],z[20]);
 Grain128_Gen #(.i(21)) GG21 (f,g,enable, fx[21],gx[21],z[21]);
 Grain128_Gen #(.i(22)) GG22 (f,g,enable, fx[22],gx[22],z[22]);
 Grain128_Gen #(.i(23)) GG23 (f,g,enable, fx[23],gx[23],z[23]);
 Grain128_Gen #(.i(24)) GG24 (f,g,enable, fx[24],gx[24],z[24]);
 Grain128_Gen #(.i(25)) GG25 (f,g,enable, fx[25],gx[25],z[25]);
 Grain128_Gen #(.i(26)) GG26 (f,g,enable, fx[26],gx[26],z[26]);
 Grain128_Gen #(.i(27)) GG27 (f,g,enable, fx[27],gx[27],z[27]);
 Grain128_Gen #(.i(28)) GG28 (f,g,enable, fx[28],gx[28],z[28]);
 Grain128_Gen #(.i(29)) GG29 (f,g,enable, fx[29],gx[29],z[29]);
 Grain128_Gen #(.i(30)) GG30 (f,g,enable, fx[30],gx[30],z[30]);
 Grain128_Gen #(.i(31)) GG31 (f,g,enable, fx[31],gx[31],z[31]);


// =========================================

endmodule





// *****************************************
//   Grain128 32  - Restricted output width
//
// *****************************************

module Grain128_32r (Key,IV,rekey,pause, keystream_r,valid, clk,aclr);
parameter w = 16, u = 32;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream_r;
 output valid;

 input  clk,aclr;

 wire [128-1:0]  f,g;

 wire [u-1:0]  fx,gx;

 wire [u-1:0] z;

 wire enable;
 
 wire [u-1:0] keystream;


// =========================================

 Grain128Reg #(.w(u)) REG0 (Key,{64'h0,IV},rekey,pause, fx,gx, f,g, clk,aclr);

 Grain128Control  #(.w(u),.j(3)) CTRL0 (rekey,pause,z, enable,keystream,valid, clk,aclr);

 Grain128_Gen #(.i(0))  GG00 (f,g,enable, fx[0], gx[0], z[0] );
 Grain128_Gen #(.i(1))  GG01 (f,g,enable, fx[1], gx[1], z[1] );
 Grain128_Gen #(.i(2))  GG02 (f,g,enable, fx[2], gx[2], z[2] );
 Grain128_Gen #(.i(3))  GG03 (f,g,enable, fx[3], gx[3], z[3] );
 Grain128_Gen #(.i(4))  GG04 (f,g,enable, fx[4], gx[4], z[4] );
 Grain128_Gen #(.i(5))  GG05 (f,g,enable, fx[5], gx[5], z[5] );
 Grain128_Gen #(.i(6))  GG06 (f,g,enable, fx[6], gx[6], z[6] );
 Grain128_Gen #(.i(7))  GG07 (f,g,enable, fx[7], gx[7], z[7] );
 Grain128_Gen #(.i(8))  GG08 (f,g,enable, fx[8], gx[8], z[8] );
 Grain128_Gen #(.i(9))  GG09 (f,g,enable, fx[9], gx[9], z[9] );
 Grain128_Gen #(.i(10)) GG10 (f,g,enable, fx[10],gx[10],z[10]);
 Grain128_Gen #(.i(11)) GG11 (f,g,enable, fx[11],gx[11],z[11]);
 Grain128_Gen #(.i(12)) GG12 (f,g,enable, fx[12],gx[12],z[12]);
 Grain128_Gen #(.i(13)) GG13 (f,g,enable, fx[13],gx[13],z[13]);
 Grain128_Gen #(.i(14)) GG14 (f,g,enable, fx[14],gx[14],z[14]);
 Grain128_Gen #(.i(15)) GG15 (f,g,enable, fx[15],gx[15],z[15]);
 Grain128_Gen #(.i(16)) GG16 (f,g,enable, fx[16],gx[16],z[16]);
 Grain128_Gen #(.i(17)) GG17 (f,g,enable, fx[17],gx[17],z[17]);
 Grain128_Gen #(.i(18)) GG18 (f,g,enable, fx[18],gx[18],z[18]);
 Grain128_Gen #(.i(19)) GG19 (f,g,enable, fx[19],gx[19],z[19]);
 Grain128_Gen #(.i(20)) GG20 (f,g,enable, fx[20],gx[20],z[20]);
 Grain128_Gen #(.i(21)) GG21 (f,g,enable, fx[21],gx[21],z[21]);
 Grain128_Gen #(.i(22)) GG22 (f,g,enable, fx[22],gx[22],z[22]);
 Grain128_Gen #(.i(23)) GG23 (f,g,enable, fx[23],gx[23],z[23]);
 Grain128_Gen #(.i(24)) GG24 (f,g,enable, fx[24],gx[24],z[24]);
 Grain128_Gen #(.i(25)) GG25 (f,g,enable, fx[25],gx[25],z[25]);
 Grain128_Gen #(.i(26)) GG26 (f,g,enable, fx[26],gx[26],z[26]);
 Grain128_Gen #(.i(27)) GG27 (f,g,enable, fx[27],gx[27],z[27]);
 Grain128_Gen #(.i(28)) GG28 (f,g,enable, fx[28],gx[28],z[28]);
 Grain128_Gen #(.i(29)) GG29 (f,g,enable, fx[29],gx[29],z[29]);
 Grain128_Gen #(.i(30)) GG30 (f,g,enable, fx[30],gx[30],z[30]);
 Grain128_Gen #(.i(31)) GG31 (f,g,enable, fx[31],gx[31],z[31]);

 assign keystream_r = keystream[w-1:0];

// =========================================

endmodule






// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Grain128 32 - Pipelined and restricted 
//                 outout size for keysearch
//
// *****************************************

module Grain128p_32r (Key,IV,rekey,pause, keystream_r,valid, clk,aclr);
parameter w = 16, u = 32;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream_r;
 output valid;

 input  clk,aclr;

 wire [u-1:0]  keystream;
 wire [128-1:0]  f0,f1,f2,f3,f4,f5,f6,f7,f8;
 wire [128-1:0]  g0,g1,g2,g3,g4,g5,g6,g7,g8; 
 wire  val0,val1,val2,val3,val4,val5,val6,val7,val8;


// =========================================

 Grain128p_Start #(.w(u)) S0  ({32'hffffffff,64'h0,IV},Key,rekey,pause, f0,g0,val0, clk,aclr);
 Grain128p_Init  #(.w(u)) I1  (f0,g0,val0,pause, f1,g1,val1, clk,aclr);
 Grain128p_Init  #(.w(u)) I2  (f1,g1,val1,pause, f2,g2,val2, clk,aclr);
 Grain128p_Init  #(.w(u)) I3  (f2,g2,val2,pause, f3,g3,val3, clk,aclr);
 Grain128p_Init  #(.w(u)) I4  (f3,g3,val3,pause, f4,g4,val4, clk,aclr);
 Grain128p_Init  #(.w(u)) I5  (f4,g4,val4,pause, f5,g5,val5, clk,aclr);
 Grain128p_Init  #(.w(u)) I6  (f5,g5,val5,pause, f6,g6,val6, clk,aclr);
 Grain128p_Init  #(.w(u)) I7  (f6,g6,val6,pause, f7,g7,val7, clk,aclr);
 Grain128p_Init  #(.w(u)) I8  (f7,g7,val7,pause, f8,g8,val8, clk,aclr);
 Grain128p_Out   #(.w(u)) O9  (f8,g8,val8,pause, keystream,valid, clk,aclr);

 assign keystream_r = keystream[w-1:0];

// =========================================

endmodule




// *****************************************
//   Grain128 16 - Pipelined for keysearch
//
// *****************************************

module Grain128p_16 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [128-1:0]  f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,f11,f12,f13,f14,f15,f16;
 wire [128-1:0]  g0,g1,g2,g3,g4,g5,g6,g7,g8,g9,g10,g11,g12,g13,g14,g15,g16; 
 wire  val0,val1,val2,val3,val4,val5,val6,val7,val8,val9,val10,val11,val12,val13,val14,val15,val16;


// =========================================

 Grain128p_Start #(.w(w)) S0  ({32'hffffffff,64'h0,IV},Key,rekey,pause, f0,g0,val0, clk,aclr);
 Grain128p_Init  #(.w(w)) I1  (f0,g0,val0,pause, f1,g1,val1, clk,aclr);
 Grain128p_Init  #(.w(w)) I2  (f1,g1,val1,pause, f2,g2,val2, clk,aclr);
 Grain128p_Init  #(.w(w)) I3  (f2,g2,val2,pause, f3,g3,val3, clk,aclr);
 Grain128p_Init  #(.w(w)) I4  (f3,g3,val3,pause, f4,g4,val4, clk,aclr);
 Grain128p_Init  #(.w(w)) I5  (f4,g4,val4,pause, f5,g5,val5, clk,aclr);
 Grain128p_Init  #(.w(w)) I6  (f5,g5,val5,pause, f6,g6,val6, clk,aclr);
 Grain128p_Init  #(.w(w)) I7  (f6,g6,val6,pause, f7,g7,val7, clk,aclr);
 Grain128p_Init  #(.w(w)) I8  (f7,g7,val7,pause, f8,g8,val8, clk,aclr);
 Grain128p_Init  #(.w(w)) I9  (f8,g8,val8,pause, f9,g9,val9, clk,aclr);
 Grain128p_Init  #(.w(w)) I10 (f9,g9,val9,pause, f10,g10,val10, clk,aclr);
 Grain128p_Init  #(.w(w)) I11 (f10,g10,val10,pause, f11,g11,val11, clk,aclr);
 Grain128p_Init  #(.w(w)) I12 (f11,g11,val11,pause, f12,g12,val12, clk,aclr);
 Grain128p_Init  #(.w(w)) I13 (f12,g12,val12,pause, f13,g13,val13, clk,aclr);
 Grain128p_Init  #(.w(w)) I14 (f13,g13,val13,pause, f14,g14,val14, clk,aclr);
 Grain128p_Init  #(.w(w)) I15 (f14,g14,val14,pause, f15,g15,val15, clk,aclr);
 Grain128p_Init  #(.w(w)) I16 (f15,g15,val15,pause, f16,g16,val16, clk,aclr); 
 Grain128p_Out   #(.w(w)) O17 (f16,g16,val16,pause, keystream,valid, clk,aclr);


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------





// *****************************************
//   Grain1 X - Compacted Dual Pipe with
//              output size restriction

//

//   note: requires verfication   
//
// *****************************************

module Grain128X_n2r (Key,IV,rekey,pause, keystream_r,valid, clk,aclr);
parameter w = 16, u = 32;

// =========================================

 input  [128-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w*2-1:0] keystream_r;
 output [2-1:0] valid;

 input  clk,aclr;

 wire [128-1:0]  f0,f1,f2,f3,f4,f5,f6,f7,f8;
 wire [128-1:0]  g0,g1,g2,g3,g4,g5,g6,g7,g8;
 wire  val0,val1,val2,val3,val4,val5,val6,val7,val8;

 wire [128-1:0]  f1x,f2x,f3x,f4x,f5x,f6x,f7x,f8x;
 wire [128-1:0]  g1x,g2x,g3x,g4x,g5x,g6x,g7x,g8x;
 wire  val3x,val4x,val5x,val6x,val7x,val8x;
 
 wire [u*2-1:0] keystream;


// =========================================

 Grain128p_Start #(.w(u)) S0  ({32'hffffffff,64'h0,IV},{1'b0,Key[126:0]},rekey,pause, f0,g0,val0, clk,aclr);

 Grain128p_Init  #(.w(u)) I1  (f0,g0,val0,pause, f1,g1,val1, clk,aclr);
 Grain128p_Init  #(.w(u)) I1x (f0,{1'b1,g0[126:0]},val0,pause, f1x,g1x, , clk,aclr); 


 Grain128p_Init  #(.w(u)) I2  (f1,g1,val1,pause,   f2,g2,val2, clk,aclr);
 //Grain128p_Init #(.w(u)) I2x (f1x,g1x,val1,pause, f2x,g2x,   , clk,aclr);   // non-compacted version, useful for debug
 Grain128p_Init  #(.w(u)) I2x (f1,{g1x[127],g1[126:96],1'b1,g1[94:0]},val1,pause, f2x,g2x,val2x, clk,aclr); 
 
 Grain128p_Init  #(.w(u)) I3  (f2,g2,val2,pause,   f3,g3,val3, clk,aclr);
//Grain128p_Init #(.w(u)) I3x (f2x,g2x,val2,pause, f3x,g3x,val3x, clk,aclr);   // non-compacted version, useful for debug
 Grain128p_Init  #(.w(u)) I3x ({f2x[127],f2[126:119],f2x[118],f2[117:103],f2x[102],f2[101:97],f2x[96],f2[95:0]},   
                              {g2x[127:126],g2[125],g2x[124:123],g2[122:119],g2x[118],g2[117:108],g2x[107],g2[106:103],g2x[102],g2[101],g2x[100],g2[99:97],g2x[96:95],g2[94:64],1'b1,g2[62:0]},
                               val2,pause, f3x,g3x,val3x, clk,aclr);
                
 Grain128p_Init  #(.w(u)) I4  (f3,g3,val3,pause,   f4,g4,val4,  clk,aclr);
//Grain128p_Init #(.w(u)) I4x (f3x,g3x,val3x,pause, f4x,g4x,val4x, clk,aclr);   // non-compacted version, useful for debug
 Grain128p_Init  #(.w(u)) I4x ({f3x[127],f3[126],f3x[125:121],f3[120],f3x[119:117],f3[116:115],f3x[114:113],f3[112],f3x[111],f3[110],f3x[109:107],f3[106],f3x[105],f3[104],f3x[103:101],f3[100],f3x[99],f3[98],f3x[97:95],f3[94:87],f3x[86],f3[85:71],f3x[70],f3[69:65],f3x[64],f3[63:0]},
                              {g3x[127:123],g3[122],g3x[121],g3[120],g3x[119:118],g3[117:115],g3x[114:111],g3[110],g3x[109:107],g3[106],g3x[105],g3[104],g3x[103:94],g3[93],g3x[92:91],g3[90:87],g3x[86],g3[85:76],g3x[75],g3[74:71],g3x[70],g3[69],g3x[68],g3[67:65],g3x[64:63],g3[62:32],1'b1,g3[30:0]},
                               val3x,pause,  f4x,g4x,val4x,  clk,aclr);
                              
 Grain128p_Init  #(.w(u)) I5  (f4,g4,val4,pause,    f5,g5,val5,  clk,aclr);
//Grain128p_Init #(.w(u)) I5x (f4x,g4x,val4x,pause, f5x,g5x,val5x, clk,aclr);   // non-compacted version, useful for debug
 Grain128p_Init  #(.w(u)) I5x ({f4x[127:95],f4[94],f4x[93:89],f4[88],f4x[87:85],f4[84:83],f4x[82:81],f4[80],f4x[79],f4[78],f4x[77:75],f4[74],f4x[73],f4[72],f4x[71:69],f4[68],f4x[67],f4[66],f4x[65:63],f4[62:55],f4x[54],f4[53:39],f4x[38],f4[37:33],f4x[32],f4[31:0]},
                              {g4x[127:91],g4[90],g4x[89],g4[88],g4x[87:86],g4[85:83],g4x[82:79],g4[78],g4x[77:75],g4[74],g4x[73],g4[72],g4x[71:62],g4[61],g4x[60:59],g4[58:55],g4x[54],g4[53:44],g4x[43],g4[42:39],g4x[38],g4[37],g4x[36],g4[35:33],g4x[32:31],g4[30:0]},
                               val4x,pause,  f5x,g5x,val5x,  clk,aclr);
                              
 Grain128p_Init  #(.w(u)) I6  (f5,g5,val5,pause,    f6,g6,val6,    clk,aclr);
//Grain128p_Init #(.w(u)) I6x (f5x,g5x,val5x,pause, f6x,g6x,val6x, clk,aclr);   // non-compacted version, useful for debug
 Grain128p_Init  #(.w(u)) I6x ({f5x[127:63],f5[62],f5x[61:57],f5[56],f5x[55:53],f5[52:51],f5x[50:49],f5[48],f5x[47],f5[46],f5x[45:43],f5[42],f5x[41],f5[40],f5x[39:37],f5[36],f5x[35],f5[34],f5x[33:31],f5[30:23],f5x[22],f5[21:7],f5x[6],f5[5:1],f5x[0]},
                              {g5x[127:59],g5[58],g5x[57],g5[56],g5x[55:54],g5[53:51],g5x[50:47],g5[46],g5x[45:43],g5[42],g5x[41],g5[40],g5x[39:30],g5[29],g5x[28:27],g5[26:23],g5x[22],g5[21:12],g5x[11],g5[10:7],g5x[6],g5[5],g5x[4],g5[3:1],g5x[0]},
                               val5x,pause,  f6x,g6x,val6x,  clk,aclr);
  
 Grain128p_Init  #(.w(u)) I7  (f6,g6,val6,pause,    f7,g7,val7,    clk,aclr);
//Grain128p_Init #(.w(u)) I7x (f6x,g6x,val6x,pause, f7x,g7x,val7x, clk,aclr);   // non-compacted version, useful for debug
 Grain128p_Init  #(.w(u)) I7x ({f6x[127:31],f6[30],f6x[29:25],f6[24],f6x[23:21],f6[20:19],f6x[18:17],f6[16],f6x[15],f6[14],f6x[13:11],f6[10],f6x[9],f6[8],f6x[7:5],f6[4],f6x[3],f6[2],f6x[1:0]},
                              {g6x[127:27],g6[26],g6x[25],g6[24],g6x[23:22],g6[21:19],g6x[18:15],g6[14],g6x[13:11],g6[10],g6x[9],g6[8],g6x[7:0]},
                               val6x,pause,  f7x,g7x,val7x,  clk,aclr);
   
 Grain128p_Init  #(.w(u)) I8  (f7,g7,val7,pause,    f8,g8,val8,   clk,aclr); 
 Grain128p_Init  #(.w(u)) I8x (f7x,g7x,val7x,pause, f8x,g8x,val8x, clk,aclr);

 Grain128p_Out   #(.w(u)) O9  (f8,g8,val8,pause,    keystream[u-1:0],    valid[0], clk,aclr);
 Grain128p_Out   #(.w(u)) O9x (f8x,g8x,val8x,pause, keystream[u*2-1:u],  valid[1], clk,aclr);

 assign keystream_r = {keystream[u*1+w-1:u*1],keystream[u*0+w-1:u*0+0]};


// =========================================

endmodule




// ------------------------------------------------------
// ------------------------------------------------------





// *****************************************
//   Grain128 - Pipeline Start Stage
//
// *****************************************

module Grain128p_Start (load_f,load_g,rekey,pause, f,g,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [128-1:0]  load_g,load_f;
 input  rekey,pause;

 output [128-1:0]  f,g;
 output valid;

 input  clk,aclr;


// =========================================


 Grain128pReg #(.w(w)) REG0 (load_f,load_g,rekey,pause, f,g, clk,aclr);

 Grain128pControl #(.w(w)) CTRL0 (rekey,pause,{(w){1'b0}},  ,valid, clk,aclr);
 

// =========================================

endmodule




// *****************************************
//   Grain 128 - Pipelined Initialise Stage
//
// *****************************************

module Grain128p_Init (f,g,rekey,pause, fReg,gReg,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [128-1:0]  f,g;
 input  rekey,pause;

 output [128-1:0]  fReg,gReg;
 output valid;

 input  clk,aclr;

 wire [w-1:0]  fx,gx;


// =========================================

 genvar x;
 generate for (x=0; x<w; x=x+1) begin: GG
  Grain128_Gen #(.i(x))  p (f,g,{1'b0}, fx[x], gx[x]      );
  end endgenerate

 Grain128pReg #(.w(w)) REG0 ({fx,f[128-1:w]},{gx,g[128-1:w]},rekey,pause, fReg,gReg, clk,aclr);

 Grain128pControl #(.w(w)) CTRL0 (rekey,pause,{(w){1'b0}},  ,valid, clk,aclr);


// =========================================

endmodule




// *****************************************
//   Grain 128 - Pipeline Output Stage
//
// *****************************************

module Grain128p_Out (f,g,rekey,pause, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [128-1:0]  f,g;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [w-1:0] z;


// =========================================


 genvar x;
 generate for (x=0; x<w; x=x+1) begin: GG
  Grain128_Gen #(.i(x))  p (f,g,{1'b1},      ,      , z[x] );
  end endgenerate


 Grain128pControl #(.w(w)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);
 

// =========================================

endmodule




// ------------------------------------------------------
// ------------------------------------------------------





// *****************************************
//   Grain 128 Pipelined - Register Setup
//
// *****************************************

module Grain128pReg (load_f,load_g,rekey,pause, f,g, clk,aclr);
parameter w = 1;

// =========================================

 input  [128-1:0]  load_f,load_g;
 input  rekey,pause;

 output reg [128-1:0] f,g;

 input  clk,aclr;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)          g <= 128'b0;
   else if(!pause) begin                 // avoids 0 propagation on pause
          if(!rekey) g <= 128'b0;
          else       g <= load_g[128-1:0];
    end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)          f <= 128'b0;
   else if(!pause) begin 
          if(!rekey) f <= 128'b0;
          else       f <= load_f[128-1:0];
    end
  end


// =========================================

endmodule




// *****************************************
//   Grain 128 Pipelined - Control Unit
//
// *****************************************

module Grain128pControl (set,pause,z, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  set,pause;

 output reg valid;

 input  clk,aclr;

 input [w-1:0] z;
 output reg [w-1:0] keystream;


// =========================================

 always@(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 1'b0;
   else if(!pause) begin            // avoids 0 propagation on pause
          if(!set) valid <= 1'b0;
          else     valid <= 1'b1;
    end
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        keystream <= {(w){1'b0}};
   else if(!pause) begin 
          if(!set) keystream <= {(w){1'b0}};
          else     keystream <= z;
    end
  end


// =========================================

endmodule

